
# This file was generated by 'versioneer.py' (0.29) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2026-02-19T05:43:47-0800",
 "dirty": false,
 "error": null,
 "full-revisionid": "3b7d8f9834eaa0362b807df5c1132d2a0bdb7b2a",
 "version": "2.6.107"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
