<!DOCTYPE html>
<html>
<head>
    <meta charset=utf-8>
    <meta name=viewport content="width=device-width,initial-scale=1">
    <link
        href="https://fonts.googleapis.com/css?family=Source+Code+Pro|Source+Code+Pro:700|Source+Sans+Pro|Source+Sans+Pro:700"
        rel=stylesheet
    >
    <title>Flame</title>

    <script type="module" crossorigin src="{{ served_app_js_url }}"></script>
    <link rel="stylesheet" href="{{ served_app_css_url }}">
</head>
<body>
  <div id="app"></div>
</body>
</html>
